#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

FILE *unfHandle;
size_t unfSize;
uint8_t *unfData;

FILE *nesHandle;
char nesName[256];

uint16_t inesMapper;
uint16_t inesSubmapper;
size_t inesPRG_ROM;
size_t inesPRG_RAM;
size_t inesPRG_NVRAM;
size_t inesPRG_Padding;
size_t inesCHR_ROM;
size_t inesCHR_RAM;
size_t inesCHR_Padding;
uint8_t inesHeader[16];
bool inesBattery;

char *unifMapper;
bool unifBattery;
size_t unifPRGsize[4], unifCHRsize[4];
uint8_t *unifPRGdata[4], *unifCHRdata[4];
enum { UnknownMirroring =-1, Horizontal =0, Vertical =1, OneScreenA =2, OneScreenB =3, ExtraMemory =4, MapperControlled =5 } unifMirroring, inesMirroring;
enum { UnknownTV =-1, NTSC =0, PAL =1, Dual =2} unifTV;

void checkWRAM8() {
	if (unifBattery) inesPRG_NVRAM =8192; else inesPRG_RAM =8192;
}
void checkWRAM16() {
	if (unifBattery) { inesPRG_NVRAM =8192; inesPRG_RAM =8192; } else inesPRG_RAM =16384;
}
void checkWRAM32() {
	if (unifBattery) inesPRG_NVRAM =32768; else inesPRG_RAM =32768;
}
void checkCHRRAM() {
	if (!inesCHR_ROM) inesCHR_RAM =8192;
}

void buildINESInfo() {
	memset(inesHeader, 0, 16);
	inesHeader[0] = 'N'; inesHeader[1] = 'E'; inesHeader[2] = 'S'; inesHeader[3] = 26;
	
	// Calculate PRG-ROM size and number of banks
	inesPRG_ROM =unifPRGsize[0] +unifPRGsize[1] +unifPRGsize[2] +unifPRGsize[3];
	inesPRG_Padding =inesPRG_ROM & 16383;
	int inesPRG_ROMbanks =inesPRG_ROM >> 14;
	if (inesPRG_Padding) inesPRG_ROMbanks++;

	// Calculate CHR-ROM size and number of banks
	inesCHR_ROM =unifCHRsize[0] +unifCHRsize[1] +unifCHRsize[2] +unifCHRsize[3];
	inesCHR_Padding =inesCHR_ROM & 8191;
	int inesCHR_ROMbanks =inesCHR_ROM >> 13;
	if (inesCHR_Padding) inesCHR_ROMbanks++;
	
	// Put PRG and CHR ROM banks into header
	inesHeader[4] = inesPRG_ROMbanks & 0xFF;
	inesHeader[5] = inesCHR_ROMbanks & 0xFF;
	inesHeader[9] = ((inesPRG_ROMbanks >>8) &0x0F) | (((inesCHR_ROMbanks >>8) &0x0F)<<4);
	
	// Process the mapper string and set up mirroring, PRG-RAM and CHR-RAM accordingly
	inesPRG_RAM =0; inesCHR_RAM =0; inesPRG_NVRAM =0; inesMirroring = unifMirroring; inesMapper =0; inesSubmapper =0;
	if (strstr(unifMapper, "NES-NROM") != NULL) { inesMapper =0; } else
	if (strstr(unifMapper, "HVC-NROM") != NULL) { inesMapper =0; } else
	if (strstr(unifMapper, "NES-HROM") != NULL) { inesMapper =0; inesMirroring =Vertical; } else
	if (strstr(unifMapper, "HVC-HROM") != NULL) { inesMapper =0; inesMirroring =Vertical; } else
	if (strstr(unifMapper, "CPROM") != NULL) { inesMapper =13; inesCHR_RAM =16*1024; } else
	if (strstr(unifMapper, "CNROM") != NULL) { inesMapper =3; inesSubmapper =2; } else
	if (strstr(unifMapper, "DEROM") != NULL) { inesMapper =206; } else
	if (strstr(unifMapper, "DE1ROM") != NULL) { inesMapper =206; } else
	if (strstr(unifMapper, "DRROM") != NULL) { inesMapper =206; inesMirroring =ExtraMemory; } else
	if (strstr(unifMapper, "EKROM") != NULL) { inesMapper =5; checkWRAM8();} else
	if (strstr(unifMapper, "ELROM") != NULL) { inesMapper =5; } else
	if (strstr(unifMapper, "ETROM") != NULL) { inesMapper =5; checkWRAM16();} else
	if (strstr(unifMapper, "EWROM") != NULL) { inesMapper =5; checkWRAM32();} else
	if (strstr(unifMapper, "FJROM") != NULL) { inesMapper =10; checkWRAM8();} else
	if (strstr(unifMapper, "FKROM") != NULL) { inesMapper =10; checkWRAM8();} else
	if (strstr(unifMapper, "GNROM") != NULL) { inesMapper =66; } else
	if (strstr(unifMapper, "MHROM") != NULL) { inesMapper =66; } else
	if (strstr(unifMapper, "UN1ROM")!= NULL) { inesMapper =94; inesSubmapper =2; inesCHR_RAM =8192; } else
	if (strstr(unifMapper, "UNROM") != NULL) { inesMapper =2; inesSubmapper =2; inesCHR_RAM =8192; } else
	if (strstr(unifMapper, "UO1ROM")!= NULL) { inesMapper =94; inesSubmapper =2; inesCHR_RAM =8192; } else
	if (strstr(unifMapper, "JLROM") != NULL) { inesMapper =69; } else
	if (strstr(unifMapper, "JSROM") != NULL) { inesMapper =69; inesPRG_RAM =8192;} else
	if (strstr(unifMapper, "BAT-E301") != NULL) { inesMapper =69; } else
	if (strstr(unifMapper, "GRM-E301") != NULL) { inesMapper =69; } else
	if (strstr(unifMapper, "HEB-E301") != NULL) { inesMapper =69; } else
	if (strstr(unifMapper, "BAT-E301") != NULL) { inesMapper =69; } else
	if (strstr(unifMapper, "PYK-E301") != NULL) { inesMapper =69; } else
	if (strstr(unifMapper, "FC-GMK") != NULL) { inesMapper =69; } else
	if (strstr(unifMapper, "FC-DBT") != NULL) { inesMapper =69; inesPRG_RAM =8192;} else
	if (strstr(unifMapper, "BBW-E301") != NULL) { inesMapper =69; inesPRG_NVRAM =8192;} else
	if (strstr(unifMapper, "DD2-E301") != NULL) { inesMapper =69; inesPRG_NVRAM =8192;} else
	if (strstr(unifMapper, "PNROM") != NULL) { inesMapper =9; } else
	if (strstr(unifMapper, "PEEOROM ") != NULL) { inesMapper =9; } else
	if (strstr(unifMapper, "SBROM") != NULL) { inesMapper =1;} else
	if (strstr(unifMapper, "SAROM") != NULL) { inesMapper =1; checkWRAM8();} else
	if (strstr(unifMapper, "SCROM") != NULL) { inesMapper =1;} else
	if (strstr(unifMapper, "SC1ROM")!= NULL) { inesMapper =1;} else
	if (strstr(unifMapper, "SEROM") != NULL) { inesMapper =1; inesSubmapper =5;} else
	if (strstr(unifMapper, "SFROM") != NULL) { inesMapper =1;} else
	if (strstr(unifMapper, "SF1ROM")!= NULL) { inesMapper =1;} else
	if (strstr(unifMapper, "SFEXPROM")!=NULL){ inesMapper =1;} else
	if (strstr(unifMapper, "SGROM") != NULL) { inesMapper =1; checkCHRRAM(); } else
	if (strstr(unifMapper, "SHROM") != NULL) { inesMapper =1; inesSubmapper =5;} else
	if (strstr(unifMapper, "SH1ROM")!= NULL) { inesMapper =1; inesSubmapper =5;} else
	if (strstr(unifMapper, "SIROM") != NULL) { inesMapper =1; checkWRAM8();} else
	if (strstr(unifMapper, "SJROM") != NULL) { inesMapper =1; checkWRAM8();} else
	if (strstr(unifMapper, "SKROM") != NULL) { inesMapper =1; checkWRAM8();} else
	if (strstr(unifMapper, "SLROM") != NULL) { inesMapper =1;} else
	if (strstr(unifMapper, "SL1ROM")!= NULL) { inesMapper =1;} else
	if (strstr(unifMapper, "SL2ROM")!= NULL) { inesMapper =1;} else
	if (strstr(unifMapper, "SL3ROM")!= NULL) { inesMapper =1;} else
	if (strstr(unifMapper, "SLRROM")!= NULL) { inesMapper =1;} else
	if (strstr(unifMapper, "SNROM") != NULL) { inesMapper =1; checkCHRRAM(); checkWRAM8();} else
	if (strstr(unifMapper, "SOROM") != NULL) { inesMapper =1; checkCHRRAM(); checkWRAM16();} else
	if (strstr(unifMapper, "SUROM") != NULL) { inesMapper =1; checkCHRRAM(); checkWRAM8();} else
	if (strstr(unifMapper, "SXROM") != NULL) { inesMapper =1; checkCHRRAM(); checkWRAM32();} else
	if (strstr(unifMapper, "TBROM") != NULL) { inesMapper =4; } else
	if (strstr(unifMapper, "TEROM") != NULL) { inesMapper =4; } else
	if (strstr(unifMapper, "TFROM") != NULL) { inesMapper =4; } else
	if (strstr(unifMapper, "TGROM") != NULL) { inesMapper =4; checkCHRRAM(); } else
	if (strstr(unifMapper, "TKROM") != NULL) { inesMapper =4; checkWRAM8(); } else
	if (strstr(unifMapper, "TK1ROM")!= NULL) { inesMapper =4; checkWRAM8(); } else
	if (strstr(unifMapper, "TKSROM")!= NULL) { inesMapper =4; checkWRAM8(); } else
	if (strstr(unifMapper, "TLROM") != NULL) { inesMapper =4; } else
	if (strstr(unifMapper, "TL1ROM")!= NULL) { inesMapper =4; } else
	if (strstr(unifMapper, "TL2ROM")!= NULL) { inesMapper =4; } else
	if (strstr(unifMapper, "TLSROM")!= NULL) { inesMapper =4; } else
	if (strstr(unifMapper, "TNROM") != NULL) { inesMapper =4; checkCHRRAM(); checkWRAM8(); } else
	if (strstr(unifMapper, "TQROM") != NULL) { inesMapper =4; inesCHR_RAM =8192; } else
	if (strstr(unifMapper, "TR1ROM")!= NULL) { inesMapper =4; inesMirroring =ExtraMemory; } else
	if (strstr(unifMapper, "TSROM") != NULL) { inesMapper =4; } else
	if (strstr(unifMapper, "TVROM") != NULL) { inesMapper =4; inesMirroring =ExtraMemory; } else
	if (strstr(unifMapper, "HKROM") != NULL) { inesMapper =4; inesSubmapper =1; inesPRG_NVRAM =1024; } else
	if (strstr(unifMapper, "22211") != NULL) { inesMapper =132; } else
	if (strstr(unifMapper, "42in1ResetSwitch") != NULL) { inesMapper = 226; } else
	if (strstr(unifMapper, "603-5052") != NULL) { inesMapper = 238; } else
	if (strstr(unifMapper, "8237") != NULL) { inesMapper = 215; } else
	if (strstr(unifMapper, "CC-21") != NULL) { inesMapper = 27; } else
	if (strstr(unifMapper, "D1038") != NULL) { inesMapper = 60; } else
	if (strstr(unifMapper, "FK23C") != NULL) { inesMapper = 176; } else
	if (strstr(unifMapper, "FS304") != NULL) { inesMapper = 172; } else
	if (strstr(unifMapper, "GK-192") != NULL) { inesMapper = 58; } else
	if (strstr(unifMapper, "H2288") != NULL) { inesMapper = 123; } else
	if (strstr(unifMapper, "KONAMI-QTAI") != NULL) { inesMapper = 190; } else
	if (strstr(unifMapper, "KS7032") != NULL) { inesMapper = 142; } else
	if (strstr(unifMapper, "KS7057") != NULL) { inesMapper = 220; } else
	if (strstr(unifMapper, "LH32") != NULL) { inesMapper = 125; } else
	if (strstr(unifMapper, "N625092") != NULL) { inesMapper = 221; } else
	if (strstr(unifMapper, "NTBROM") != NULL) { inesMapper = 68; } else
	if (strstr(unifMapper, "RET-CUFROM") != NULL) { inesMapper = 29; inesCHR_RAM =32*1024; } else
	if (strstr(unifMapper, "SA-002") != NULL) { inesMapper = 136; } else
	if (strstr(unifMapper, "SA-0036") != NULL) { inesMapper = 149; } else
	if (strstr(unifMapper, "SA-0037") != NULL) { inesMapper = 148; } else
	if (strstr(unifMapper, "SA-009") != NULL) { inesMapper = 79; } else	// FCEUX uses Mapper 160!
	if (strstr(unifMapper, "SA-016-1M") != NULL) { inesMapper = 79; } else	// FCEU uses Mapper 146!
	if (strstr(unifMapper, "SA-72007") != NULL) { inesMapper = 145; } else
	if (strstr(unifMapper, "SA-72008") != NULL) { inesMapper = 133; } else
	if (strstr(unifMapper, "SA-9602B") != NULL) { inesMapper = 513; inesCHR_RAM =32*1024; } else
	if (strstr(unifMapper, "SA-NROM") != NULL) { inesMapper = 143; } else
	if (strstr(unifMapper, "SC-127") != NULL) { inesMapper = 35; } else
	if (strstr(unifMapper, "SL12") != NULL) { inesMapper = 116; } else
	if (strstr(unifMapper, "SL1632") != NULL) { inesMapper = 14; } else
	if (strstr(unifMapper, "Sachen-74LS374N") != NULL) { inesMapper = 150; } else
	if (strstr(unifMapper, "Sachen-74LS374NA") != NULL) { inesMapper = 243; } else
	if (strstr(unifMapper, "Sachen-8259A") != NULL) { inesMapper = 141; } else
	if (strstr(unifMapper, "Sachen-8259B") != NULL) { inesMapper = 138; } else
	if (strstr(unifMapper, "Sachen-8259C") != NULL) { inesMapper = 139; } else
	if (strstr(unifMapper, "Sachen-8259D") != NULL) { inesMapper = 137; } else
	if (strstr(unifMapper, "SMB2J") != NULL) { inesMapper = 43; } else
	if (strstr(unifMapper, "SuperHIK8in1") != NULL) { inesMapper = 45; } else
	if (strstr(unifMapper, "Supervision16in1") != NULL) { inesMapper = 53; } else
	if (strstr(unifMapper, "TC-U01-1.5M") != NULL) { inesMapper = 147; } else
	if (strstr(unifMapper, "TEK90") != NULL) { inesMapper = 90; } else
	if (strstr(unifMapper, "DANCE") != NULL) { inesMapper = 256; } else
	if (strstr(unifMapper, "OneBus") != NULL) { inesMapper = 256; } else
	if (strstr(unifMapper, "PEC-586") != NULL) { inesMapper = 257; } else
	if (strstr(unifMapper, "158B") != NULL) { inesMapper = 258; } else
	if (strstr(unifMapper, "F-15") != NULL) { inesMapper = 259; } else
	if (strstr(unifMapper, "NTD-03") != NULL) { inesMapper = 260; } else
	if (strstr(unifMapper, "11160") != NULL) { inesMapper = 261; } else
	if (strstr(unifMapper, "SHERO") != NULL) { inesMapper = 262; inesMirroring =ExtraMemory; inesCHR_RAM =8192; } else
	if (strstr(unifMapper, "KOF97") != NULL) { inesMapper = 263; } else
	if (strstr(unifMapper, "YOKO") != NULL) { inesMapper = 264; } else
	if (strstr(unifMapper, "VRC7") != NULL) { inesMapper = 265; } else
	if (strstr(unifMapper, "CITYFIGHT") != NULL) { inesMapper = 266; } else
	if (strstr(unifMapper, "HROM") !=NULL) { inesMapper =0; inesMirroring =Vertical; } else
	fprintf(stderr, "%s: Unknown mapper name\n", unifMapper);
	
	if (!inesCHR_ROM && !inesCHR_RAM) inesCHR_RAM = 8192;
	inesBattery = unifBattery;
	if (inesPRG_NVRAM) inesBattery = true;
	inesHeader[6] = ((inesMapper & 0x0F)<<4) |
			((inesMirroring ==ExtraMemory)? 0x08: 0x00) |
			(inesBattery? 0x02: 0x00) |
			((inesMirroring == Vertical)? 0x01: 0x00);
	inesHeader[7] = 0x08 | (inesMapper & 0xF0); // 1000b = NES 2.0
	inesHeader[8] = (inesSubmapper <<4) | (inesMapper >> 8);
	
	int PRG_RAMshift = 0; inesPRG_RAM >>= 7; while (inesPRG_RAM) { PRG_RAMshift++; inesPRG_RAM >>= 1; }
	int PRG_NVRAMshift = 0; inesPRG_NVRAM >>= 7; while (inesPRG_NVRAM) { PRG_NVRAMshift++; inesPRG_NVRAM >>= 1; }
	inesHeader[10] = (PRG_NVRAMshift <<4) | PRG_RAMshift;
	
	int CHR_RAMshift = 0; inesCHR_RAM >>= 7; while (inesCHR_RAM) { CHR_RAMshift++; inesCHR_RAM >>= 1; }
	inesHeader[11] = CHR_RAMshift;
	
	if (unifTV >0) inesHeader[12] = unifTV;
}

void processFile(const char* unfName) {
	strcpy(nesName, unfName);
	bool foundExtension =false;
	for (int i =0; i <(strlen(nesName) -3); i++)  {
		if (nesName[i+0]=='.' && 
		   (nesName[i+1]=='u' || nesName[i+1]=='U') && 
		   (nesName[i+2]=='n' || nesName[i+2]=='N') && 
		   (nesName[i+3]=='f' || nesName[i+3]=='F') &&
		    nesName[i+4]=='\0') {
			nesName[i+1] = 'n';
			nesName[i+2] = 'e';
			nesName[i+3] = 's';
			foundExtension =true;
		}
	}
	printf ("%s\n", nesName);
	if (!foundExtension) {
		fprintf(stderr, "%s: Does not have .UNF extension\n", unfName);
		return;
	}
	
	if (unfHandle =fopen(unfName, "rb")) {
		fseek(unfHandle, 0, SEEK_END);
		unfSize =ftell(unfHandle);
		if (unfData =malloc(unfSize)) {
			fseek(unfHandle, 0, SEEK_SET);
			fread(unfData, 1, unfSize, unfHandle);
			if (memcmp(unfData, "UNIF", 4)==0) {
				// Reset UNIF flags
				for (int i=0; i<4; i++) {
					unifPRGsize[i] =0;
					unifCHRsize[i] =0;
					unifPRGdata[i] =NULL;
					unifCHRdata[i] =NULL;
				}
				unifMirroring = UnknownMirroring;
				unifBattery = false;
				unifMapper = NULL;
				unifTV = UnknownTV;
				
				// Parse chunks
				uint8_t* s=unfData +32;
				while ((s +8) <= (unfData +unfSize)) {
					size_t chunkSize = (s[7]<<24) +(s[6]<<16) +(s[5]<<8) +s[4];
					uint8_t* chunkStart = s+8;
					if (memcmp(s, "MAPR", 4)==0) unifMapper =chunkStart; else
					if (memcmp(s, "TVCI", 4)==0) unifTV =*chunkStart; else
					if (memcmp(s, "MIRR", 4)==0) unifMirroring =*chunkStart; else
					if (memcmp(s, "BATR", 4)==0) unifBattery =*chunkStart; else
					if (memcmp(s, "PRG0", 4)==0) { unifPRGsize[0] =chunkSize; unifPRGdata[0] =chunkStart; } else
					if (memcmp(s, "PRG1", 4)==0) { unifPRGsize[1] =chunkSize; unifPRGdata[1] =chunkStart; } else
					if (memcmp(s, "PRG2", 4)==0) { unifPRGsize[2] =chunkSize; unifPRGdata[2] =chunkStart; } else
					if (memcmp(s, "PRG3", 4)==0) { unifPRGsize[3] =chunkSize; unifPRGdata[3] =chunkStart; } else
					if (memcmp(s, "CHR0", 4)==0) { unifCHRsize[0] =chunkSize; unifCHRdata[0] =chunkStart; } else
					if (memcmp(s, "CHR1", 4)==0) { unifCHRsize[1] =chunkSize; unifCHRdata[1] =chunkStart; } else
					if (memcmp(s, "CHR2", 4)==0) { unifCHRsize[2] =chunkSize; unifCHRdata[2] =chunkStart; } else
					if (memcmp(s, "CHR3", 4)==0) { unifCHRsize[3] =chunkSize; unifCHRdata[3] =chunkStart; }
					s = chunkStart +chunkSize;
				}
				buildINESInfo();
				if (nesHandle =fopen(nesName, "wb")) {
					fwrite(inesHeader, 1, 16, nesHandle);
					if (unifPRGsize[0]) fwrite(unifPRGdata[0], 1, unifPRGsize[0], nesHandle);
					if (unifPRGsize[1]) fwrite(unifPRGdata[1], 1, unifPRGsize[1], nesHandle);
					if (unifPRGsize[2]) fwrite(unifPRGdata[2], 1, unifPRGsize[2], nesHandle);
					if (unifPRGsize[3]) fwrite(unifPRGdata[3], 1, unifPRGsize[3], nesHandle);
					if (unifCHRsize[0]) fwrite(unifCHRdata[0], 1, unifCHRsize[0], nesHandle);
					if (unifCHRsize[1]) fwrite(unifCHRdata[1], 1, unifCHRsize[1], nesHandle);
					if (unifCHRsize[2]) fwrite(unifCHRdata[2], 1, unifCHRsize[2], nesHandle);
					if (unifCHRsize[3]) fwrite(unifCHRdata[3], 1, unifCHRsize[3], nesHandle);
					fclose(nesHandle);
				} else perror(nesName);
			} else fprintf(stderr, "%s: UNIF signature not found in header\n", unfName);
			free(unfData);
		} else fprintf(stderr, "%s: Cannot allocate %u bytes\n", unfName, unfSize);
		fclose(unfHandle);
	} else perror(unfName);
}

int main(int argc, char **argv) {
	for (int arg =1; arg <argc; arg++) processFile(argv[arg]);
	return EXIT_SUCCESS;
}